#include "tmesh.h"

#include <fstream>
#include <iostream>
#include <float.h>
#include <math.h>

TMesh::TMesh() : verts(0), vertsN(0), tris(0), trisN(0), 
  colors(0),  normals(0) {};

void TMesh::Load(char *fname) {

  ifstream ifs(fname, ios::binary);
  if (ifs.fail()) {
    cerr << "INFO: cannot open file: " << fname << endl;
    return;
  }

  ifs.read((char*)&vertsN, sizeof(int));
  char yn;
  ifs.read(&yn, 1); // always xyz
  if (yn != 'y') {
    cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
    return;
  }
  if (verts)
    delete verts;
  verts = new V3[vertsN];

  ifs.read(&yn, 1); // colors 3 floats
  if (colors)
    delete colors;
  colors = 0;
  if (yn == 'y') {
    colors = new V3[vertsN];
  }

  ifs.read(&yn, 1); // normals 3 floats
  if (normals)
    delete normals;
  normals = 0;
  if (yn == 'y') {
    normals = new V3[vertsN];
  }

  ifs.read(&yn, 1); // texture coordinates 2 floats
  float *tcs = 0; // don't have texture coordinates for now
  if (tcs)
    delete tcs;
  tcs = 0;
  if (yn == 'y') {
    tcs = new float[vertsN*2];
  }

  ifs.read((char*)verts, vertsN*3*sizeof(float)); // load verts

  if (colors) {
    ifs.read((char*)colors, vertsN*3*sizeof(float)); // load colors
  }

  if (normals)
    ifs.read((char*)normals, vertsN*3*sizeof(float)); // load normals

  if (tcs)
    ifs.read((char*)tcs, vertsN*2*sizeof(float)); // load texture coordinates

  ifs.read((char*)&trisN, sizeof(int));
  if (tris)
    delete tris;
  tris = new unsigned int[trisN*3];
  ifs.read((char*)tris, trisN*3*sizeof(unsigned int)); // read tiangles

  ifs.close();

  cerr << "INFO: loaded " << vertsN << " verts, " << trisN << " tris from " << endl << "      " << fname << endl;
  cerr << "      xyz " << ((colors) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;


}

void TMesh::RenderWF(FrameBuffer *fb, PPC *ppc, 
                     unsigned int wfc) {

  V3 *projVerts = new V3[vertsN];
  for (int i = 0; i < vertsN; i++) {
    ppc->Project(verts[i], projVerts[i]);
  }

  for (int i = 0; i < trisN; i++) {
    for (int j = 0; j < 3; j++) {
      int _j = (j+1)%3;
      int vi0  = tris[3*i+j];
      int vi1  = tris[3*i+_j];
      V3 c0, c1;
      if (colors) {
        c0 = colors[vi0];
        c1 = colors[vi1];
      }
      else {
        c0.SetFromColor(wfc);
        c1 = c0;
      }
      fb->DrawSegment2D(projVerts[vi0], projVerts[vi1], c0, c1);
    }
  }

  delete projVerts;

}

void TMesh::Translate(V3 trans) {

  for (int i = 0; i < vertsN; i++) {
    verts[i] = verts[i] + trans;
  }

}

V3 TMesh::GetCenter() {

  V3 center(0.0f, 0.0f, 0.0f);
  for (int i = 0; i < vertsN; i++) {
    center = center + verts[i];
  }
  center = center / (float) vertsN;
  return center;

}

AABB TMesh::GetAABB() {

  AABB ret;
  for (int i = 0; i < vertsN; i++) {
    ret.AddPoint(verts[i]);
  }

  return ret;

}

void TMesh::Scale(float diag) {

  AABB aabb = GetAABB();
  float currDiag = (aabb[0]-aabb[1]).Length();
  float scalef = diag / currDiag;
  V3 c = (aabb[0]+aabb[1]) / 2.0f;
  for (int i = 0; i < vertsN; i++) {
    verts[i] = c + (verts[i]-c)*scalef;
  }

}

void TMesh::SetCenter(V3 newCenter) {

  V3 oldCenter = GetCenter();
  Translate(newCenter-oldCenter);

}

void TMesh::RenderFilled(FrameBuffer *fb, 
                         PPC *ppc, unsigned int meshColor) {

  
  V3 *projVerts = new V3[vertsN];
  for (int i = 0; i < vertsN; i++) {
    ppc->Project(verts[i], projVerts[i]);
  }

  V3 color;
  color.SetFromColor(meshColor);

  for (int i = 0; i < trisN; i++) {
    int vi[3]= {tris[3*i+0], tris[3*i+1], tris[3*i+2]};

    // any vertex behind eye
    if (
      projVerts[vi[0]][0] == FLT_MAX ||
      projVerts[vi[1]][0] == FLT_MAX ||
      projVerts[vi[2]][0] == FLT_MAX) {
      continue;
    }
    AABB projaabb;
    projaabb.AddPoint(projVerts[vi[0]]);
    projaabb.AddPoint(projVerts[vi[1]]);
    projaabb.AddPoint(projVerts[vi[2]]);

    // bbox of proj tri outside of image frame
    if (
      projaabb[0][0] > (float) fb->w ||
      projaabb[1][0] < 0.0f ||
      projaabb[0][1] > (float) fb->h ||
      projaabb[1][1] < 0.0f)
      continue;

    // clip bbox to image frame
    if (projaabb[0][0] < 0.0f)
      projaabb[0][0] = 0.0f;
    if (projaabb[1][0] > (float) fb->w)
      projaabb[1][0] = (float) fb->w;
    if (projaabb[0][1] < 0.0f)
      projaabb[0][1] = 0.0f;
    if (projaabb[1][1] > (float) fb->h)
      projaabb[1][1] = (float) fb->h;


    // placeholder color
    if (colors) {
      color = colors[vi[0]];
    }

    // edge equation setup
    M3x3 eeqs;
    for (int j = 0; j < 3; j++) {
      int _j = (j+1)%3;
      V3 p = projVerts[vi[j]];
      V3 q = projVerts[vi[_j]];
      V3 r = projVerts[vi[(j+2)%3]];
      eeqs[j][0] = q[1]-p[1];
      eeqs[j][1] = p[0]-q[0];
      eeqs[j][2] = -p[1]*eeqs[j][1]-p[0]*eeqs[j][0];
      if (eeqs[j]*V3(r[0], r[1], 1.0f) < 0.0f) {
        eeqs[j] = eeqs[j]*-1.0f;
      }
    }

    // rasterization parameter setup
    M3x3 ssim;
    ssim[0] = V3(projVerts[vi[0]]); ssim[0][2] = 1.0f;
    ssim[1] = V3(projVerts[vi[1]]); ssim[1][2] = 1.0f;
    ssim[2] = V3(projVerts[vi[2]]); ssim[2][2] = 1.0f;
    ssim = ssim.Invert();
    V3 ABCz = ssim * V3(projVerts[vi[0]][2], 
      projVerts[vi[1]][2], projVerts[vi[2]][2]);

    V3 shades[3];
    float ka = 0.1f;
    V3 lightv(0.0f, 1.0f, 0.0f);
    lightv = lightv.Normalized();
    V3 objcolor(1.0f, 0.0f, 0.0f);
    for (int i = 0; i < 3; i++) {
      float kd = normals[vi[i]]*lightv;
      kd = (kd < 0.0f) ? 0.0f : kd;
      float k = ka + kd*(1-ka);
      shades[i] = objcolor*k;
    }

#if 0
    V3 ABCred = ssim * V3(colors[vi[0]][0], 
      colors[vi[1]][0], colors[vi[2]][0]);
    V3 ABCgreen = ssim * V3(colors[vi[0]][1], 
      colors[vi[1]][1], colors[vi[2]][1]);
    V3 ABCblue = ssim * V3(colors[vi[0]][2], 
      colors[vi[1]][2], colors[vi[2]][2]);
#else
    V3 ABCred = ssim * V3(shades[0][0], 
      shades[1][0], shades[2][0]);
    V3 ABCgreen = ssim * V3(shades[0][1], 
      shades[1][1], shades[2][1]);
    V3 ABCblue = ssim * V3(shades[0][2], 
      shades[1][2], shades[2][2]);
#endif

    M3x3 ABCcolors;
    ABCcolors[0] = ABCred;
    ABCcolors[1] = ABCgreen;
    ABCcolors[2] = ABCblue;

    int top = (int) floorf(.5f+projaabb[0][1]);
    int left = (int) floorf(.5f+projaabb[0][0]);
    int down = (int) floorf(-.5f+projaabb[1][1]);
    int right = (int) floorf(-.5f+projaabb[1][0]);
    for (int v = top; v <= down; v++) {
      for (int u = left; u <= right; u++) {
        V3 pixv(.5f+(float)u, .5f+(float)v, 1.0f);
        if (
          eeqs[0]*pixv < 0.0f || 
          eeqs[1]*pixv < 0.0f || 
          eeqs[2]*pixv < 0.0f)
          continue;
        float currz = ABCz*pixv;
        if (fb->IsCloserThenSet(u, v, currz)) {
          V3 currColor = ABCcolors*pixv;
          fb->Set(u, v, currColor.GetColor());
        }
      }
    }

  }

  delete []projVerts;

}
